/*
 * Copyright (c) 1999 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Java Apache 
 *    Project. <http://java.apache.org/>"
 *
 * 4. The names "Java Apache Element Construction Set", "Java Apache ECS" and 
 *    "Java Apache Project" must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Products derived from this software may not be called 
 *    "Java Apache Element Construction Set" nor "Java Apache ECS" appear 
 *    in their names without prior written permission of the 
 *    Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Java Apache 
 *    Project. <http://java.apache.org/>"
 *    
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Project. For more information
 * on the Java Apache Project please see <http://java.apache.org/>.
 *
 */
package org.apache.ecs;

import java.util.Enumeration;
import java.util.Hashtable;
import java.io.PrintWriter;
import java.io.OutputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;


/**
    This class implements the ElementFactory.

    @version $Id: Element.java,v 1.1.1.1 1999/04/20 01:18:57 jonbolt Exp $
    @author <a href="mailto:snagy@servletapi.com">Stephan Nagy</a>
    @author <a href="mailto:jon@clearink.com">Jon S. Stevens</a>
*/
public abstract class Element implements ElementFactory, AttributeRegistry
{

    /** Filter to use to escape input */
    private Filter filter = new CharacterFilter();

    /** Filter to use to escape attribute input */
    
    private Filter attribute_filter = filter;   // By default the attribute filter and the element filter are the same.

    /** Should we filter the value of &lt;&gt;VALUE&lt;/&gt; */
    private boolean filter_state = true;

    /** Should we filter the value of the element attributes */
    private boolean filter_attribute_state = false;

    /** Version number for codebase */
    private static final String VERSION = "0.01A";

    /** Default value to set case type */
    private int case_type = MIXEDCASE;

    /** Where to get the element definition from */
    private Hashtable element = new Hashtable();

    /** Does this element need a closing tag? */
    private boolean end_element = true;

    /** Does this element attribute value need a =""? */
    protected static final String NO_ATTRIBUTE_VALUE = "HTMLGEN_NO_ATTRIBUTE_VALUE";

    /** What codeset are we going to use the default is 8859_1 */
    private String codeset = "8859_1";

    /** What the start modifier should be */
    private String start_modifier = new String();

    /** What the end modifier should be */
    private String end_modifier = new String();

    /** What the modifier should be...optimization variable */
    private boolean start_modifier_defined = false;

    /** What the modifier should be...optimization variable */
    private boolean end_modifier_defined = false;

    /** Base class for defining all elements. */
    public Element()
    {
    }

    /** Get the Hashtable representation of an element. */
    protected Hashtable getElementHashEntry()
    {
        return(element);
    }

    /**
        Set the element modifier the default is no modifier.
        @param modifier the modifier for the element a ? would result <br>
        in &lt;? ?&gt;
    */
    public Element setModifier(String modifier)
    {
        setStartModifier(modifier);
        setEndModifier(modifier);
        setStartModifierDefined(true);
        setEndModifierDefined(true);
        return(this);
    }

    /**
        Set the element modifier the default is no modifier.
        @param modifier the modifier for the element a ? would result <br>
        in &lt;? &gt;
    */
    public Element setStartModifier(String modifier)
    {
        this.start_modifier = modifier;
        setStartModifierDefined(true);
        return(this);
    }

    /**
        Set the element modifier the default is no modifier.
        @param modifier the modifier for the element a ? would result <br>
        in &lt; ?&gt;
    */
    public Element setEndModifier(String modifier)
    {
        this.end_modifier = modifier;
        setEndModifierDefined(true);
        return(this);
    }

    /**
        Find out if we want to filter the elements attributes or not.
    */
    protected boolean getAttributeFilterState()
    {
        return(filter_attribute_state);
    }
    
    /**
        Tell the element if we want to filter its attriubtes.
        @param filter_attribute_state do we want to filter the attributes of this element?
    */
    public Element setAttributeFilterState(boolean filter_attribute_state)
    {
        this.filter_attribute_state = filter_attribute_state;
        return(this);
    }

    /**
        Set up a new filter for all element attributes.
        @param Filter the filter we want to use for element attributes.  By <br>
        default it is the same as is used for the value of the tag. It is assumed<br>
        that if you create a new filter you must want to use it.
    */
    public Element setAttributeFilter(Filter attribute_filter)
    {
        filter_attribute_state = true; // If your setting up a filter you must want to filter.
        this.attribute_filter = attribute_filter;
        return(this);
    }

    /**
        Set the element modifier the default is no modifier.
        @param modifier the modifier for the element a ? would result <br>
        in &lt;? ?&gt;
    */
    public Element setModifier(String start_modifier,String end_modifier)
    {
        this.start_modifier = start_modifier;
        this.end_modifier = end_modifier;
        setStartModifierDefined(true);
        setEndModifierDefined(true);
        return(this);
    }
    /**
        What is the element modifier for this tag.
    */
    public String getStartModifier()
    {
        return(start_modifier);
    }

    /**
        What is the element modifier for this tag.
    */
    public String getEndModifier()
    {
        return(end_modifier);
    }

    /**
        Remove the element modifier for this tag.
    */
    public Element removeStartModifier()
    {
        setStartModifierDefined(false);
        return(this);
    }

    /**
        Remove the element modifier for this tag.
    */
    public Element removeEndModifier()
    {
        setEndModifierDefined(false);
        return(this);
    }

    /**
        Modifier optimization
    */
    private void setStartModifierDefined(boolean val)
    {
        this.start_modifier_defined = val;
    }

    /**
        Modifier optimization
    */
    private void setEndModifierDefined(boolean val)
    {
        this.end_modifier_defined = val;
    }

    /**
        Modifier optimization
    */
    private boolean getStartModifierDefined()
    {
        return this.start_modifier_defined;
    }

    /**
        Modifier optimization
    */
    private boolean getEndModifierDefined()
    {
        return this.end_modifier_defined;
    }

    /** Add an attribute to the element. */
    public void addAttribute(String attribute_name, Object attribute_value)
    {
        element.put(attribute_name, attribute_value);
    }

    public void removeAttribute(String attribute_name)
    {
        try
        {
            element.remove(attribute_name);
        }
        catch ( Exception e )
        {
        }
    }

    public boolean hasAttribute(String attribute)
    {
        return(element.containsKey(attribute));
    }

    /** Override the default filter with a user supplied filter. */
    public void setFilter(Filter filter)
    {
        this.filter = filter;
    }

    /**
        set the filter state of the element.
        @param filter_state (true|false) do we want to use a filter? default true;
    */
    public void setFilterState(boolean filter_state)
    {
        this.filter_state = filter_state;
    }

    /**
        get the filter state of the element.
    */
    public boolean getFilterState()
    {
        return(filter_state);
    }
    
    /** Get the filter that is currently in use. */
    public Filter getFilter()
    {
        return(filter);
    }
    
    /** What version of htmlGen is this? */
    public String getVersion()
    {
        return(VERSION);
    }

    /** Determine what case the html element should be rendered. Default is MIXEDCASE */
    public void setCase(int case_type)
    {
    this.case_type = case_type;
    }

    /** Find out what case the element is supposed to be rendered in.*/
    public int getCase()
    {
        return(case_type);
    }

    /** Determine what type of element this is A,BR,EM,ACRONYM */
    public void setHtmlElementType(String html_element)
    {
        element.put("_tag",html_element);
    }

    /** Find out what type of element this is. */
    public String getHtmlElementType()
    {
        return((String) element.get("_tag"));
    }

    /** Determine what codeset should be used. */
    public void setCodeSet(String codeset)
    {
        this.codeset = codeset;
    }

    /** Find out what the current codeset is. */
    public String getCodeSet()
    {
        return(codeset);
    }

    /** Does this element need a closing tag. Default is true */
    public void setNeedClosingTag(boolean close)
    {
        end_element = close;
    }

    /** Determine if this element needs to be closed. */
    public boolean getNeedClosingTag()
    {
        return(end_element);
    }

    /**     Alter the case of this element so that it is rendered according
            to whatever the value of case_type is.  Default is MIXEDCASE */
    protected String alterCase(String value)
    {
        switch (getCase())
        {
            case UPPERCASE:
                return value.toUpperCase();
            case LOWERCASE:
                return value.toLowerCase();
            default:
                return value;
        }
    }

    /** Override toString() method to print something meaningful */
    public final String toString()
    {
       String out = new String();
       try
        {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            BufferedOutputStream bos = new BufferedOutputStream(baos);
            output(bos);
            bos.flush();
            out = baos.toString();
            bos.close();
            baos.close();
        }
        catch (IOException ioe)
        {
        }
        return(out);
    }

    /** Override toString(encode) method to print something meaningful */
    public final String toString(String codeset)
    {

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        BufferedOutputStream bos = new BufferedOutputStream(baos);
        String out = new String();
        try
        {
            output(bos);
            bos.flush();
            out = baos.toString(codeset);  // XXXXX:  there has got to be a better way.
            bos.close();
            baos.close();
        }
        catch (UnsupportedEncodingException use)
        {
        }
        catch (IOException ioe)
        {
        }
        return(out);
    }

    /** Add the element to the designated OutputStream. */
    public void output(OutputStream out)
    {

        Hashtable ce = (Hashtable) getElementHashEntry().clone();
        String html_content = new String();

        try
        {

            String html_tag = (String) ce.get("_tag");
            ce.remove("_tag"); // get rid of the tag
            html_content = (String) ce.get("_text");
            ce.remove("_text"); // get rid of the _text attribute
            
            out.write('<');

            if(getStartModifierDefined())
            {
                out.write(getStartModifier().getBytes());
            }

            out.write(alterCase(html_tag).getBytes());
            Enumeration enum = ce.keys();
            String value = new String(); // avoid creating a new string object on each pass through the loop

            while (enum.hasMoreElements())
            {
                String tag = (String) enum.nextElement();
                if(getAttributeFilterState())
                {
                    value = filter.process((String) ce.get(tag));
                }
                else
                {
                    value = (String) ce.get(tag);
                }
                out.write(' ');
                out.write(alterCase(tag).getBytes());
                if(!value.equalsIgnoreCase(NO_ATTRIBUTE_VALUE))
                {
                    out.write('=');
                    out.write('"');
                    out.write(value.getBytes());
                    out.write('"');
                }
            }

            if(getEndModifierDefined())
            {
                out.write(getEndModifier().getBytes());
            }
            out.write('>');
            out.write(html_content.getBytes());

            // Not all multi-part elements need a closing tag make it optional
            if (getNeedClosingTag())
            {
                out.write('<');
                out.write('/');
                out.write(alterCase(html_tag).getBytes());
                out.write('>');
            }

        }
        catch (IOException ioe)
        {
        }
    }

    /** Add element to the designated PrintWriter. */
    public void output(PrintWriter pw)
    {
        Hashtable ce = (Hashtable) getElementHashEntry().clone();
        String html_content = new String();

        String html_tag = (String) ce.get("_tag");
        ce.remove("_tag"); // get rid of the tag
        html_content = (String) ce.get("_text");
        ce.remove("_text"); // get rid of the _text attribute
        pw.write('<');

        if(getStartModifierDefined())
        {
            pw.write(getStartModifier());
        }

        pw.write(alterCase(html_tag));
        Enumeration enum = ce.keys();
        String value = new String();

        while (enum.hasMoreElements())
        {

            String tag = (String) enum.nextElement();
            if(getAttributeFilterState())
            {
                value = filter.process((String) ce.get(tag));
            }
            else
            {
                value = (String) ce.get(tag);
            }
            pw.write(' ');
            pw.write(alterCase(tag));
            if(!value.equalsIgnoreCase(NO_ATTRIBUTE_VALUE))
            {
                pw.write('=');
                pw.write('"');
                pw.write(value);
                pw.write('"');
            }
        }

        if(getEndModifierDefined())
        {
            pw.write(getEndModifier());
        }

        pw.write('>');
        pw.write(html_content);
        if(getNeedClosingTag())
        {
            pw.write('<');
            pw.write('/');
            pw.write(alterCase(html_tag));
            pw.write('>');
        }
    }
}